from __future__ import annotations

from sqlalchemy import (
    Double,
    Index,
    Integer,
    String,
    Text,
)
from sqlalchemy.orm import declarative_base

from diracx.db.sql.utils.types import SmarterDateTime

from ..utils import Column, EnumBackedBool, NullColumn

PilotAgentsDBBase = declarative_base()


class PilotAgents(PilotAgentsDBBase):
    __tablename__ = "PilotAgents"

    pilot_id = Column("PilotID", Integer, autoincrement=True, primary_key=True)
    initial_job_id = Column("InitialJobID", Integer, default=0)
    current_job_id = Column("CurrentJobID", Integer, default=0)
    pilot_job_reference = Column("PilotJobReference", String(255), default="Unknown")
    pilot_stamp = Column("PilotStamp", String(32), default="")
    destination_site = Column("DestinationSite", String(128), default="NotAssigned")
    queue = Column("Queue", String(128), default="Unknown")
    grid_site = Column("GridSite", String(128), default="Unknown")
    vo = Column("VO", String(128))
    grid_type = Column("GridType", String(32), default="LCG")
    benchmark = Column("BenchMark", Double, default=0.0)
    submission_time = NullColumn("SubmissionTime", SmarterDateTime)
    last_update_time = NullColumn("LastUpdateTime", SmarterDateTime)
    status = Column("Status", String(32), default="Unknown")
    status_reason = Column("StatusReason", String(255), default="Unknown")
    accounting_sent = Column("AccountingSent", EnumBackedBool(), default=False)

    __table_args__ = (
        Index("PilotJobReference", "PilotJobReference"),
        Index("Status", "Status"),
        Index("Statuskey", "GridSite", "DestinationSite", "Status"),
    )


class JobToPilotMapping(PilotAgentsDBBase):
    __tablename__ = "JobToPilotMapping"

    pilot_id = Column("PilotID", Integer, primary_key=True)
    job_id = Column("JobID", Integer, primary_key=True)
    start_time = Column("StartTime", SmarterDateTime)

    __table_args__ = (Index("JobID", "JobID"), Index("PilotID", "PilotID"))


class PilotOutput(PilotAgentsDBBase):
    __tablename__ = "PilotOutput"

    pilot_id = Column("PilotID", Integer, primary_key=True)
    std_output = Column("StdOutput", Text)
    std_error = Column("StdError", Text)



from __future__ import annotations

__all__ = (
    "AuthDB",
    "JobDB",
    "JobLoggingDB",
    "PilotAgentsDB",
    "SandboxMetadataDB",
    "TaskQueueDB",
)

from .auth.db import AuthDB
from .job.db import JobDB
from .job_logging.db import JobLoggingDB
from .pilot_agents.db import PilotAgentsDB
from .sandbox_metadata.db import SandboxMetadataDB
from .task_queue.db import TaskQueueDB



from __future__ import annotations

from datetime import UTC, datetime

from sqlalchemy import Integer, Numeric, PrimaryKeyConstraint, String, TypeDecorator
from sqlalchemy.orm import declarative_base

from ..utils import Column, DateNowColumn

JobLoggingDBBase = declarative_base()


class MagicEpochDateTime(TypeDecorator):
    """A SQLAlchemy type that stores a datetime as a numeric value representing the
    seconds elapsed since MAGIC_EPOC_NUMBER. The underlying column is defined as
    Numeric(12,3) which provides a fixed-precision representation.
    """

    impl = Numeric(12, 3)
    cache_ok = True

    MAGIC_EPOC_NUMBER = 1270000000

    def process_bind_param(self, value, dialect):
        """Convert a Python datetime to a numeric value: (timestamp - MAGIC_EPOC_NUMBER).
        The result is rounded to three decimal places.
        """
        if value is None:
            return None
        if isinstance(value, datetime):
            # Convert datetime to seconds since the Unix epoch, subtract our magic epoch,
            # and round to three decimal places.
            epoch_seconds = (
                value.replace(tzinfo=UTC).timestamp() - self.MAGIC_EPOC_NUMBER
            )
            return round(epoch_seconds, 3)
        raise ValueError(
            "Expected a datetime object for MagicEpochDateTime bind parameter."
        )

    def process_result_value(self, value, dialect):
        """Convert the numeric database value back into a Python datetime by reversing the
        stored difference (adding MAGIC_EPOC_NUMBER).
        """
        if value is None:
            return None
        # Carefully convert from Decimal to datetime to avoid losing precision
        value += self.MAGIC_EPOC_NUMBER
        value_int = int(value)
        result = datetime.fromtimestamp(value_int, tz=UTC)
        return result.replace(microsecond=int((value - value_int) * 1_000_000))


class LoggingInfo(JobLoggingDBBase):
    __tablename__ = "LoggingInfo"
    job_id = Column("JobID", Integer)
    seq_num = Column("SeqNum", Integer)
    status = Column("Status", String(32), default="")
    minor_status = Column("MinorStatus", String(128), default="")
    application_status = Column("ApplicationStatus", String(255), default="")
    status_time = DateNowColumn("StatusTime")
    status_time_order = Column("StatusTimeOrder", MagicEpochDateTime, default=0)
    source = Column("StatusSource", String(32), default="Unknown")
    __table_args__ = (PrimaryKeyConstraint("JobID", "SeqNum"),)



from __future__ import annotations

import re
from datetime import UTC, datetime, timedelta, timezone

import freezegun
import pytest
import sqlalchemy
from sqlalchemy.ext.asyncio import create_async_engine

from diracx.testing.time import julian_date, mock_sqlite_time

RE_SQLITE_TIME = re.compile(r"(\d{4})-(\d{2})-(\d{2})(?: (\d{2}):(\d{2}):(\d{2}))?")


@pytest.mark.parametrize(
    "dt, expected",
    [
        (datetime(2000, 1, 1, 18, 0, 0, tzinfo=UTC), 2451545.25),
        (datetime(2000, 1, 1, 6, 0, 0, tzinfo=UTC), 2451544.75),
    ],
)
def test_julian_date(dt, expected):
    """Test the julian_date function with known values."""
    assert julian_date(dt) == expected


def test_julian_date_non_utc():
    """Test the julian_date function with a non-UTC timezone."""
    non_utc = timezone(timedelta(hours=-5))
    dt = datetime(2000, 1, 1, 18, 0, 0, tzinfo=non_utc)
    jd = julian_date(dt)
    # dt in UTC is 2000-01-01 23:00:00, so fractional day is (23-12)/24 = 11/24
    expected = 2451545 + 11 / 24
    assert abs(jd - expected) < 1e-6


@pytest.mark.parametrize("with_mock", [True, False])
async def test_freeze_sqlite_datetime(with_mock):
    """Test the SQLite DATETIME() function with freezegun."""
    engine = create_async_engine("sqlite+aiosqlite:///:memory:", future=True, echo=True)
    if with_mock:
        sqlalchemy.event.listen(engine.sync_engine, "connect", mock_sqlite_time)

    async with engine.begin() as conn:
        # DATETIME()
        result = await conn.execute(sqlalchemy.text("SELECT DATETIME()"))
        value = datetime.strptime(result.scalar_one(), "%Y-%m-%d %H:%M:%S").replace(
            tzinfo=UTC
        )
        actual = datetime.now(UTC)
        assert abs(value - actual) < timedelta(seconds=1)

        if with_mock:
            with freezegun.freeze_time("2000-10-01 12:00:00"):
                result = await conn.execute(sqlalchemy.text("SELECT DATETIME()"))
                value = datetime.strptime(
                    result.scalar_one(), "%Y-%m-%d %H:%M:%S"
                ).replace(tzinfo=UTC)
                actual = datetime(2000, 10, 1, 12, 0, 0, tzinfo=UTC)
                assert abs(value - actual) < timedelta(seconds=1)

        # DATETIME('now')
        result = await conn.execute(sqlalchemy.text("SELECT DATETIME('now')"))
        value = datetime.strptime(result.scalar_one(), "%Y-%m-%d %H:%M:%S").replace(
            tzinfo=UTC
        )
        actual = datetime.now(UTC)
        assert abs(value - actual) < timedelta(seconds=1)

        if with_mock:
            with freezegun.freeze_time("2000-10-01 12:00:00"):
                result = await conn.execute(sqlalchemy.text("SELECT DATETIME('now')"))
                value = datetime.strptime(
                    result.scalar_one(), "%Y-%m-%d %H:%M:%S"
                ).replace(tzinfo=UTC)
                actual = datetime(2000, 10, 1, 12, 0, 0, tzinfo=UTC)
                assert abs(value - actual) < timedelta(seconds=1)


@pytest.mark.parametrize("with_mock", [True, False])
async def test_freeze_sqlite_julianday(with_mock):
    """Test the SQLite JULIANDAY() function with freezegun."""
    engine = create_async_engine("sqlite+aiosqlite:///:memory:", future=True, echo=True)
    if with_mock:
        sqlalchemy.event.listen(engine.sync_engine, "connect", mock_sqlite_time)

    async with engine.begin() as conn:
        # JULIANDAY()
        result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY()"))
        actual = julian_date(datetime.now(UTC))
        assert abs(result.scalar_one() - actual) < 1e-3

        if with_mock:
            with freezegun.freeze_time("2000-10-01 12:00:00"):
                result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY()"))
                actual = julian_date(datetime(2000, 10, 1, 12, 0, 0, tzinfo=UTC))
                assert abs(result.scalar_one() - actual) < 1e-3

        # JULIANDAY('now')
        result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY('now')"))
        actual = julian_date(datetime.now(UTC))
        assert abs(result.scalar_one() - actual) < 1e-3

        if with_mock:
            with freezegun.freeze_time("2000-10-01 12:00:00"):
                result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY('now')"))
                actual = julian_date(datetime(2000, 10, 1, 12, 0, 0, tzinfo=UTC))
                assert abs(result.scalar_one() - actual) < 1e-3

        # JULIANDAY('YYYY-MM-DD')
        result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY('1996-05-30')"))
        assert result.scalar_one() == 2450233.5

        if with_mock:
            with freezegun.freeze_time("2020-12-13 12:34:56"):
                result = await conn.execute(
                    sqlalchemy.text("SELECT JULIANDAY('1996-05-30')")
                )
                assert result.scalar_one() == 2450233.5

        # JULIANDAY('YYYY-MM-DD HH:MM:SS')
        result = await conn.execute(
            sqlalchemy.text("SELECT JULIANDAY('2000-10-01 12:00:00')")
        )
        assert result.scalar_one() == 2451819.0

        if with_mock:
            with freezegun.freeze_time("2020-12-13 12:34:56"):
                result = await conn.execute(
                    sqlalchemy.text("SELECT JULIANDAY('2000-10-01 12:00:00')")
                )
                assert result.scalar_one() == 2451819.0

        # JULIANDAY('1356')
        result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY('1356')"))
        assert result.scalar_one() == 1356.0

        if with_mock:
            with freezegun.freeze_time("2020-12-13 12:34:56"):
                result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY('1356')"))
                assert result.scalar_one() == 1356.0

        # JULIANDAY('1356.12')
        result = await conn.execute(sqlalchemy.text("SELECT JULIANDAY('1356.12')"))
        assert result.scalar_one() == 1356.12

        if with_mock:
            with freezegun.freeze_time("2020-12-13 12:34:56"):
                result = await conn.execute(
                    sqlalchemy.text("SELECT JULIANDAY('1356.12')")
                )
                assert result.scalar_one() == 1356.12



from __future__ import annotations

__all__ = [
    "_get_columns",
    "utcnow",
    "Column",
    "NullColumn",
    "DateNowColumn",
    "BaseSQLDB",
    "EnumBackedBool",
    "EnumColumn",
    "apply_search_filters",
    "apply_sort_constraints",
    "substract_date",
    "hash",
    "SQLDBUnavailableError",
    "uuid7_from_datetime",
    "uuid7_to_datetime",
]

from .base import (
    BaseSQLDB,
    SQLDBUnavailableError,
    _get_columns,
    apply_search_filters,
    apply_sort_constraints,
    uuid7_from_datetime,
    uuid7_to_datetime,
)
from .functions import hash, substract_date, utcnow
from .types import Column, DateNowColumn, EnumBackedBool, EnumColumn, NullColumn



from __future__ import annotations

from datetime import datetime
from zoneinfo import ZoneInfo

import pytest
import sqlalchemy
from sqlalchemy.exc import IntegrityError

from diracx.core.exceptions import InvalidQueryError
from diracx.core.models import (
    ScalarSearchOperator,
    ScalarSearchSpec,
    SortDirection,
    SortSpec,
    VectorSearchOperator,
    VectorSearchSpec,
)
from diracx.db.sql.job.db import JobDB


@pytest.fixture
async def job_db(tmp_path):
    job_db = JobDB("sqlite+aiosqlite:///:memory:")
    async with job_db.engine_context():
        async with job_db.engine.begin() as conn:
            # set PRAGMA foreign_keys=ON if sqlite
            if job_db._db_url.startswith("sqlite"):
                await conn.exec_driver_sql("PRAGMA foreign_keys=ON")
            await conn.run_sync(job_db.metadata.create_all)
        yield job_db


@pytest.fixture
async def populated_job_db(job_db):
    """Populate the in-memory JobDB with 100 jobs using DAL calls."""
    async with job_db as db:
        jobs_to_insert = {}
        # Insert 100 jobs directly via the DAL.
        for i in range(100):
            compressed_jdl = f"CompressedJDL{i}"
            job_id = await db.create_job(compressed_jdl)
            jobs_to_insert[job_id] = {
                "JobID": job_id,
                "Status": "New",
                "Owner": f"owner{i}",
                "OwnerGroup": "owner_group1" if i < 50 else "owner_group2",
                "VO": "lhcb",
            }
        await db.insert_job_attributes(jobs_to_insert)
    yield job_db


async def test_bad_naive_datetime_used(populated_job_db):
    async with populated_job_db as db:
        compressed_jdl = "CompressedJDL0001111BadJob"
        job_id = await db.create_job(compressed_jdl)
        jobs_to_insert = {}
        jobs_to_insert[job_id] = {
            "JobID": job_id,
            "Status": "New",
            "Owner": "owner0101010101",
            "OwnerGroup": "owner_group1",
            "VO": "lhcb",
            "HeartBeatTime": datetime.now().replace(tzinfo=None),  # noqa
        }
        with pytest.raises(sqlalchemy.exc.StatementError):
            await db.insert_job_attributes(jobs_to_insert)  # should complain


async def test_timezone_converted_back_to_utc(populated_job_db):
    current_utc_dt = datetime.now(tz=ZoneInfo("UTC"))

    async with populated_job_db as db:
        compressed_jdl = "CompressedJDL0001111BadJob"
        job_id = await db.create_job(compressed_jdl)
        jobs_to_insert = {}
        jobs_to_insert[job_id] = {
            "JobID": job_id,
            "Status": "New",
            "Owner": "owner0101010101",
            "OwnerGroup": "owner_group1",
            "VO": "lhcb",
            "HeartBeatTime": current_utc_dt.astimezone(ZoneInfo("Asia/Tokyo")),
        }
        await db.insert_job_attributes(jobs_to_insert)

        total, result = await db.search(
            ["JobID", "HeartBeatTime"],
            [
                ScalarSearchSpec(
                    parameter="JobID",
                    operator=ScalarSearchOperator.EQUAL,
                    value=int(job_id),
                )
            ],
            [],
        )
        assert total == 1
        assert result[0]["HeartBeatTime"] == current_utc_dt, result


async def test_search_parameters(populated_job_db):
    """Test that we can search specific parameters for jobs in the database."""
    async with populated_job_db as job_db:
        # Search a specific parameter: JobID
        total, result = await job_db.search(["JobID"], [], [])
        assert total == 100
        assert result
        for r in result:
            assert r.keys() == {"JobID"}

        # Search a specific parameter: Status
        total, result = await job_db.search(["Status"], [], [])
        assert total == 100
        assert result
        for r in result:
            assert r.keys() == {"Status"}

        # Search for multiple parameters: JobID, Status
        total, result = await job_db.search(["JobID", "Status"], [], [])
        assert total == 100
        assert result
        for r in result:
            assert r.keys() == {"JobID", "Status"}

        # Search for a specific parameter but use distinct: Status
        total, result = await job_db.search(["Status"], [], [], distinct=True)
        assert total == 1
        assert result

        # Search for a non-existent parameter: Dummy
        with pytest.raises(InvalidQueryError):
            total, result = await job_db.search(["Dummy"], [], [])


async def test_search_conditions(populated_job_db):
    """Test that we can search for specific jobs in the database."""
    async with populated_job_db as job_db:
        # Search a specific scalar condition: JobID eq 3
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=3
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 1
        assert result
        assert len(result) == 1
        assert result[0]["JobID"] == 3

        # Search a specific scalar condition: JobID lt 3
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.LESS_THAN, value=3
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 2
        assert result
        assert len(result) == 2
        assert result[0]["JobID"] == 1
        assert result[1]["JobID"] == 2

        # Search a specific scalar condition: JobID neq 3
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.NOT_EQUAL, value=3
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 99
        assert result
        assert len(result) == 99
        assert all(r["JobID"] != 3 for r in result)

        # Search a specific scalar condition: JobID eq 5873 (does not exist)
        condition = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=5873
        )
        total, result = await job_db.search([], [condition], [])
        assert not result

        # Search a specific vector condition: JobID in 1,2,3
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[1, 2, 3]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 3
        assert result
        assert len(result) == 3
        assert all(r["JobID"] in [1, 2, 3] for r in result)

        # Search a specific vector condition: JobID in 1,2,5873 (one of them does not exist)
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[1, 2, 5873]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 2
        assert result
        assert len(result) == 2
        assert all(r["JobID"] in [1, 2] for r in result)

        # Search a specific vector condition: JobID not in 1,2,3
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.NOT_IN, values=[1, 2, 3]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 97
        assert result
        assert len(result) == 97
        assert all(r["JobID"] not in [1, 2, 3] for r in result)

        # Search a specific vector condition: JobID not in 1,2,5873 (one of them does not exist)
        condition = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.NOT_IN, values=[1, 2, 5873]
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 98
        assert result
        assert len(result) == 98
        assert all(r["JobID"] not in [1, 2] for r in result)

        # Search for multiple conditions based on different parameters: JobID eq 70, JobID in 4,5,6
        condition1 = ScalarSearchSpec(
            parameter="Owner", operator=ScalarSearchOperator.EQUAL, value="owner4"
        )
        condition2 = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[4, 5, 6]
        )
        total, result = await job_db.search([], [condition1, condition2], [])
        assert total == 1
        assert result
        assert len(result) == 1
        assert result[0]["JobID"] == 5
        assert result[0]["Owner"] == "owner4"

        # Search for multiple conditions based on the same parameter: JobID eq 70, JobID in 4,5,6
        condition1 = ScalarSearchSpec(
            parameter="JobID", operator=ScalarSearchOperator.EQUAL, value=70
        )
        condition2 = VectorSearchSpec(
            parameter="JobID", operator=VectorSearchOperator.IN, values=[4, 5, 6]
        )
        total, result = await job_db.search([], [condition1, condition2], [])
        assert total == 0
        assert not result

        # Search for a specific scalar condition: Owner not like 'owner1%'
        condition = ScalarSearchSpec(
            parameter="Owner", operator=ScalarSearchOperator.NOT_LIKE, value="owner1%"
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 100 - 11
        assert result
        assert len(result) == 100 - 11
        assert all(not r["Owner"].startswith("owner1") for r in result)

        # Search for a specific scalar condition: OwnerGroup not like 'owner_group2'
        condition = ScalarSearchSpec(
            parameter="OwnerGroup",
            operator=ScalarSearchOperator.NOT_LIKE,
            value="owner_group2",
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 100 - 50
        assert result
        assert len(result) == 100 - 50
        assert all(not r["OwnerGroup"] == "owner_group2" for r in result)

        # Search for a specific scalar condition: Owner regex '^owner\d+$'
        condition = ScalarSearchSpec(
            parameter="Owner", operator=ScalarSearchOperator.REGEX, value="^owner\\d+$"
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 100
        assert result
        assert len(result) == 100

        # Search for a specific scalar condition: JobID regex 'owner[0-3]+'
        # owner0, owner1, owner2, owner3 (4 jobs)
        # owner11 -> owner39 (30 jobs)
        condition = ScalarSearchSpec(
            parameter="Owner", operator=ScalarSearchOperator.REGEX, value="owner[0-3]+"
        )
        total, result = await job_db.search([], [condition], [])
        assert total == 34
        assert result
        assert len(result) == 34

        # Search for a specific scalar condition: JobID regex 'owner[1-'
        condition = ScalarSearchSpec(
            parameter="Owner", operator=ScalarSearchOperator.REGEX, value="owner[1-"
        )
        with pytest.raises(InvalidQueryError):
            await job_db.search([], [condition], [])


async def test_search_sorts(populated_job_db):
    """Test that we can search for jobs in the database and sort the results."""
    async with populated_job_db as job_db:
        # Search and sort by JobID in ascending order
        sort = SortSpec(parameter="JobID", direction=SortDirection.ASC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        for i, r in enumerate(result):
            assert r["JobID"] == i + 1

        # Search and sort by JobID in descending order
        sort = SortSpec(parameter="JobID", direction=SortDirection.DESC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        for i, r in enumerate(result):
            assert r["JobID"] == 100 - i

        # Search and sort by Owner in ascending order
        sort = SortSpec(parameter="Owner", direction=SortDirection.ASC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        # Assert that owner10 is before owner2 because of the lexicographical order
        assert result[2]["Owner"] == "owner10"
        assert result[12]["Owner"] == "owner2"

        # Search and sort by Owner in descending order
        sort = SortSpec(parameter="Owner", direction=SortDirection.DESC)
        total, result = await job_db.search([], [], [sort])
        assert total == 100
        assert result
        # Assert that owner10 is before owner2 because of the lexicographical order
        assert result[97]["Owner"] == "owner10"
        assert result[87]["Owner"] == "owner2"

        # Search and sort by OwnerGroup in ascending order and JobID in descending order
        sort1 = SortSpec(parameter="OwnerGroup", direction=SortDirection.ASC)
        sort2 = SortSpec(parameter="JobID", direction=SortDirection.DESC)
        total, result = await job_db.search([], [], [sort1, sort2])
        assert total == 100
        assert result
        assert result[0]["OwnerGroup"] == "owner_group1"
        assert result[0]["JobID"] == 50
        assert result[99]["OwnerGroup"] == "owner_group2"
        assert result[99]["JobID"] == 51


async def test_search_pagination(populated_job_db):
    """Test that we can search for jobs in the database."""
    async with populated_job_db as job_db:
        # Search for the first 10 jobs
        total, result = await job_db.search([], [], [], per_page=10, page=1)
        assert total == 100
        assert result
        assert len(result) == 10
        assert result[0]["JobID"] == 1

        # Search for the second 10 jobs
        total, result = await job_db.search([], [], [], per_page=10, page=2)
        assert total == 100
        assert result
        assert len(result) == 10
        assert result[0]["JobID"] == 11

        # Search for the last 10 jobs
        total, result = await job_db.search([], [], [], per_page=10, page=10)
        assert total == 100
        assert result
        assert len(result) == 10
        assert result[0]["JobID"] == 91

        # Search for the second 50 jobs
        total, result = await job_db.search([], [], [], per_page=50, page=2)
        assert total == 100
        assert result
        assert len(result) == 50
        assert result[0]["JobID"] == 51

        # Invalid page number
        total, result = await job_db.search([], [], [], per_page=10, page=11)
        assert total == 100
        assert not result

        # Invalid page number
        with pytest.raises(InvalidQueryError):
            result = await job_db.search([], [], [], per_page=10, page=0)

        # Invalid per_page number
        with pytest.raises(InvalidQueryError):
            result = await job_db.search([], [], [], per_page=0, page=1)


async def test_set_job_commands_invalid_job_id(job_db: JobDB):
    """Test that setting a command for a non-existent job raises JobNotFound."""
    async with job_db as job_db:
        with pytest.raises(IntegrityError):
            await job_db.set_job_commands([(123456, "test_command", "")])



from __future__ import annotations

import asyncio
import secrets
from datetime import datetime

import pytest
import sqlalchemy

from diracx.core.exceptions import SandboxAlreadyInsertedError, SandboxNotFoundError
from diracx.core.models import SandboxInfo, UserInfo
from diracx.db.sql.sandbox_metadata.db import SandboxMetadataDB
from diracx.db.sql.sandbox_metadata.schema import SandBoxes, SBEntityMapping


@pytest.fixture
async def sandbox_metadata_db(tmp_path):
    sandbox_metadata_db = SandboxMetadataDB("sqlite+aiosqlite:///:memory:")
    async with sandbox_metadata_db.engine_context():
        async with sandbox_metadata_db.engine.begin() as conn:
            await conn.run_sync(sandbox_metadata_db.metadata.create_all)
        yield sandbox_metadata_db


def test_get_pfn(sandbox_metadata_db: SandboxMetadataDB):
    user_info = UserInfo(
        sub="vo:sub", preferred_username="user1", dirac_group="group1", vo="vo"
    )
    sandbox_info = SandboxInfo(
        checksum="90e0ba6763c91a905bb9fd6e025aac1952ae742e6d756a31a0963aa7df7cd7b1",
        checksum_algorithm="sha256",
        format="tar.bz2",
        size=100,
    )
    pfn = sandbox_metadata_db.get_pfn("bucket1", user_info, sandbox_info)
    assert pfn == (
        "/S3/bucket1/vo/group1/user1/"
        "sha256:90e0ba6763c91a905bb9fd6e025aac1952ae742e6d756a31a0963aa7df7cd7b1.tar.bz2"
    )


async def test_insert_sandbox(sandbox_metadata_db: SandboxMetadataDB):
    # TODO: DAL tests should be very simple, such complex tests should be handled in diracx-routers
    user_info = UserInfo(
        sub="vo:sub", preferred_username="user1", dirac_group="group1", vo="vo"
    )
    pfn1 = secrets.token_hex()

    # Make sure the sandbox doesn't already exist
    db_contents = await _dump_db(sandbox_metadata_db)
    assert pfn1 not in db_contents
    async with sandbox_metadata_db:
        with pytest.raises(SandboxNotFoundError):
            await sandbox_metadata_db.sandbox_is_assigned(pfn1, "SandboxSE")

    # Insert owner
    async with sandbox_metadata_db:
        owner_id = await sandbox_metadata_db.insert_owner(user_info)
    assert owner_id == 1

    # Insert the sandbox
    async with sandbox_metadata_db:
        await sandbox_metadata_db.insert_sandbox(owner_id, "SandboxSE", pfn1, 100)
    db_contents = await _dump_db(sandbox_metadata_db)
    owner_id1, last_access_time1 = db_contents[pfn1]

    await asyncio.sleep(1)  # The timestamp only has second precision
    async with sandbox_metadata_db:
        with pytest.raises(SandboxAlreadyInsertedError):
            await sandbox_metadata_db.insert_sandbox(owner_id, "SandboxSE", pfn1, 100)

        await sandbox_metadata_db.update_sandbox_last_access_time("SandboxSE", pfn1)

    db_contents = await _dump_db(sandbox_metadata_db)
    owner_id2, last_access_time2 = db_contents[pfn1]
    assert owner_id1 == owner_id2
    assert last_access_time2 > last_access_time1

    # The sandbox still hasn't been assigned
    async with sandbox_metadata_db:
        assert not await sandbox_metadata_db.sandbox_is_assigned(pfn1, "SandboxSE")

    # Inserting again should update the last access time
    await asyncio.sleep(1)  # The timestamp only has second precision
    last_access_time3 = (await _dump_db(sandbox_metadata_db))[pfn1][1]
    assert last_access_time2 == last_access_time3
    async with sandbox_metadata_db:
        await sandbox_metadata_db.update_sandbox_last_access_time("SandboxSE", pfn1)
    last_access_time4 = (await _dump_db(sandbox_metadata_db))[pfn1][1]
    assert last_access_time2 < last_access_time4


async def _dump_db(
    sandbox_metadata_db: SandboxMetadataDB,
) -> dict[str, tuple[int, datetime]]:
    """Dump the contents of the sandbox metadata database.

    Returns a dict[pfn: str, (owner_id: int, last_access_time: datetime)]
    """
    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(
            SandBoxes.SEPFN, SandBoxes.OwnerId, SandBoxes.LastAccessTime
        )
        res = await sandbox_metadata_db.conn.execute(stmt)
        return {row.SEPFN: (row.OwnerId, row.LastAccessTime) for row in res}


async def test_assign_and_unsassign_sandbox_to_jobs(
    sandbox_metadata_db: SandboxMetadataDB,
):
    # TODO: DAL tests should be very simple, such complex tests should be handled in diracx-routers
    pfn = secrets.token_hex()
    user_info = UserInfo(
        sub="vo:sub", preferred_username="user1", dirac_group="group1", vo="vo"
    )
    dummy_jobid = 666
    sandbox_se = "SandboxSE"
    # Insert the sandbox
    async with sandbox_metadata_db:
        owner_id = await sandbox_metadata_db.insert_owner(user_info)
        await sandbox_metadata_db.insert_sandbox(owner_id, sandbox_se, pfn, 100)

    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(SandBoxes.SBId, SandBoxes.SEPFN)
        res = await sandbox_metadata_db.conn.execute(stmt)
    db_contents = {row.SEPFN: row.SBId for row in res}
    sb_id_1 = db_contents[pfn]
    # The sandbox still hasn't been assigned
    async with sandbox_metadata_db:
        assert not await sandbox_metadata_db.sandbox_is_assigned(pfn, sandbox_se)

    # Check there is no mapping
    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(
            SBEntityMapping.SBId, SBEntityMapping.EntityId, SBEntityMapping.Type
        )
        res = await sandbox_metadata_db.conn.execute(stmt)
    db_contents = {row.SBId: (row.EntityId, row.Type) for row in res}
    assert db_contents == {}

    # Assign sandbox with dummy jobid
    async with sandbox_metadata_db:
        await sandbox_metadata_db.assign_sandbox_to_jobs(
            jobs_ids=[dummy_jobid], pfn=pfn, sb_type="Output", se_name=sandbox_se
        )
    # Check if sandbox and job are mapped
    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(
            SBEntityMapping.SBId, SBEntityMapping.EntityId, SBEntityMapping.Type
        )
        res = await sandbox_metadata_db.conn.execute(stmt)
    db_contents = {row.SBId: (row.EntityId, row.Type) for row in res}

    entity_id_1, sb_type = db_contents[sb_id_1]
    assert entity_id_1 == f"Job:{dummy_jobid}"
    assert sb_type == "Output"

    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(SandBoxes.SBId, SandBoxes.SEPFN)
        res = await sandbox_metadata_db.conn.execute(stmt)
    db_contents = {row.SEPFN: row.SBId for row in res}
    sb_id_1 = db_contents[pfn]
    # The sandbox should be assigned
    async with sandbox_metadata_db:
        assert await sandbox_metadata_db.sandbox_is_assigned(pfn, sandbox_se)

    # Unassign the sandbox to job
    async with sandbox_metadata_db:
        await sandbox_metadata_db.unassign_sandboxes_to_jobs([dummy_jobid])

    # Entity should not exists anymore
    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(SBEntityMapping.SBId).where(
            SBEntityMapping.EntityId == entity_id_1
        )
        res = await sandbox_metadata_db.conn.execute(stmt)
    entity_sb_id = [row.SBId for row in res]
    assert entity_sb_id == []

    # Should not be assigned anymore
    async with sandbox_metadata_db:
        assert await sandbox_metadata_db.sandbox_is_assigned(pfn, sandbox_se) is False
    # Check the mapping has been deleted
    async with sandbox_metadata_db:
        stmt = sqlalchemy.select(SBEntityMapping.SBId)
        res = await sandbox_metadata_db.conn.execute(stmt)
    res_sb_id = [row.SBId for row in res]
    assert sb_id_1 not in res_sb_id


async def test_get_sandbox_owner_id(sandbox_metadata_db: SandboxMetadataDB):
    user_info = UserInfo(
        sub="vo:sub", preferred_username="user1", dirac_group="group1", vo="vo"
    )
    pfn = secrets.token_hex()
    sandbox_se = "SandboxSE"
    # Insert the sandbox
    async with sandbox_metadata_db:
        owner_id = await sandbox_metadata_db.insert_owner(user_info)
        await sandbox_metadata_db.insert_sandbox(owner_id, sandbox_se, pfn, 100)

    async with sandbox_metadata_db:
        sb_owner_id = await sandbox_metadata_db.get_sandbox_owner_id(pfn, sandbox_se)

    assert owner_id == 1
    assert sb_owner_id == owner_id

    async with sandbox_metadata_db:
        sb_owner_id = await sandbox_metadata_db.get_sandbox_owner_id(
            "not_found", sandbox_se
        )
    assert sb_owner_id is None



from __future__ import annotations

import logging
from contextlib import asynccontextmanager
from functools import partial
from typing import Any, AsyncGenerator

from sqlalchemy import (
    BigInteger,
    Column,
    Executable,
    MetaData,
    Table,
    and_,
    delete,
    exists,
    insert,
    literal,
    or_,
    select,
    update,
)
from sqlalchemy.exc import IntegrityError, NoResultFound

from diracx.core.exceptions import (
    SandboxAlreadyAssignedError,
    SandboxAlreadyInsertedError,
    SandboxNotFoundError,
)
from diracx.core.models import SandboxInfo, SandboxType, UserInfo
from diracx.db.sql.utils.base import BaseSQLDB
from diracx.db.sql.utils.functions import days_since, utcnow

from .schema import Base as SandboxMetadataDBBase
from .schema import SandBoxes, SBEntityMapping, SBOwners

logger = logging.getLogger(__name__)


class SandboxMetadataDB(BaseSQLDB):
    metadata = SandboxMetadataDBBase.metadata

    # Temporary table to store the sandboxes to delete, see `select_and_delete_expired`
    _temp_table = Table(
        "sb_to_delete",
        MetaData(),
        Column("SBId", BigInteger, primary_key=True),
        prefixes=["TEMPORARY"],
    )

    async def get_owner_id(self, user: UserInfo) -> int | None:
        """Get the id of the owner from the database."""
        stmt = select(SBOwners.OwnerID).where(
            SBOwners.Owner == user.preferred_username,
            SBOwners.OwnerGroup == user.dirac_group,
            SBOwners.VO == user.vo,
        )
        return (await self.conn.execute(stmt)).scalar_one_or_none()

    async def get_sandbox_owner_id(self, pfn: str, se_name: str) -> int | None:
        """Get the id of the owner of a sandbox."""
        stmt = select(SBOwners.OwnerID).where(
            SBOwners.OwnerID == SandBoxes.OwnerId,
            SandBoxes.SEName == se_name,
            SandBoxes.SEPFN == pfn,
        )
        return (await self.conn.execute(stmt)).scalar_one_or_none()

    async def insert_owner(self, user: UserInfo) -> int:
        stmt = insert(SBOwners).values(
            Owner=user.preferred_username,
            OwnerGroup=user.dirac_group,
            VO=user.vo,
        )
        result = await self.conn.execute(stmt)
        return result.lastrowid

    @staticmethod
    def get_pfn(bucket_name: str, user: UserInfo, sandbox_info: SandboxInfo) -> str:
        """Get the sandbox's user namespaced and content addressed PFN."""
        parts = [
            "S3",
            bucket_name,
            user.vo,
            user.dirac_group,
            user.preferred_username,
            f"{sandbox_info.checksum_algorithm}:{sandbox_info.checksum}.{sandbox_info.format}",
        ]
        return "/" + "/".join(parts)

    async def insert_sandbox(
        self, owner_id: int, se_name: str, pfn: str, size: int
    ) -> None:
        """Add a new sandbox in SandboxMetadataDB."""
        stmt = insert(SandBoxes).values(
            OwnerId=owner_id,
            SEName=se_name,
            SEPFN=pfn,
            Bytes=size,
            RegistrationTime=utcnow(),
            LastAccessTime=utcnow(),
        )
        try:
            await self.conn.execute(stmt)
        except IntegrityError as e:
            raise SandboxAlreadyInsertedError(pfn, se_name) from e

    async def update_sandbox_last_access_time(self, se_name: str, pfn: str) -> None:
        stmt = (
            update(SandBoxes)
            .where(SandBoxes.SEName == se_name, SandBoxes.SEPFN == pfn)
            .values(LastAccessTime=utcnow())
        )
        result = await self.conn.execute(stmt)
        if result.rowcount == 0:
            # If the update didn't affect any row, the sandbox doesn't exist
            raise SandboxNotFoundError(pfn, se_name)
        elif result.rowcount != 1:
            raise NotImplementedError(
                "More than one sandbox was updated. This should not happen."
            )

    async def sandbox_is_assigned(self, pfn: str, se_name: str) -> bool | None:
        """Checks if a sandbox exists and has been assigned."""
        stmt: Executable = select(SandBoxes.Assigned).where(
            SandBoxes.SEName == se_name, SandBoxes.SEPFN == pfn
        )
        result = await self.conn.execute(stmt)
        try:
            is_assigned = result.scalar_one()
        except NoResultFound as e:
            raise SandboxNotFoundError(pfn, se_name) from e

        return is_assigned

    @staticmethod
    def jobid_to_entity_id(job_id: int) -> str:
        """Define the entity id as 'Entity:entity_id' due to the DB definition."""
        return f"Job:{job_id}"

    async def get_sandbox_assigned_to_job(
        self, job_id: int, sb_type: SandboxType
    ) -> list[Any]:
        """Get the sandbox assign to job."""
        entity_id = self.jobid_to_entity_id(job_id)
        stmt = (
            select(SandBoxes.SEPFN)
            .where(SandBoxes.SBId == SBEntityMapping.SBId)
            .where(
                SBEntityMapping.EntityId == entity_id,
                SBEntityMapping.Type == sb_type,
            )
        )
        result = await self.conn.execute(stmt)
        return [result.scalar()]

    async def assign_sandbox_to_jobs(
        self,
        jobs_ids: list[int],
        pfn: str,
        sb_type: SandboxType,
        se_name: str,
    ) -> None:
        """Map sandbox and jobs."""
        for job_id in jobs_ids:
            # Define the entity id as 'Entity:entity_id' due to the DB definition:
            entity_id = self.jobid_to_entity_id(job_id)
            select_sb_id = select(
                SandBoxes.SBId,
                literal(entity_id).label("EntityId"),
                literal(sb_type).label("Type"),
            ).where(
                SandBoxes.SEName == se_name,
                SandBoxes.SEPFN == pfn,
            )
            stmt = insert(SBEntityMapping).from_select(
                ["SBId", "EntityId", "Type"], select_sb_id
            )
            try:
                await self.conn.execute(stmt)
            except IntegrityError as e:
                raise SandboxAlreadyAssignedError(pfn, se_name) from e

            stmt = update(SandBoxes).where(SandBoxes.SEPFN == pfn).values(Assigned=True)
            result = await self.conn.execute(stmt)
            if result.rowcount == 0:
                # If the update didn't affect any row, the sandbox doesn't exist
                # It means the previous insert didn't have any effect
                raise SandboxNotFoundError(pfn, se_name)

            assert result.rowcount == 1

    async def unassign_sandboxes_to_jobs(self, jobs_ids: list[int]) -> None:
        """Delete mapping between jobs and sandboxes."""
        for job_id in jobs_ids:
            entity_id = self.jobid_to_entity_id(job_id)
            sb_sel_stmt = select(SandBoxes.SBId)
            sb_sel_stmt = sb_sel_stmt.join(
                SBEntityMapping, SBEntityMapping.SBId == SandBoxes.SBId
            )
            sb_sel_stmt = sb_sel_stmt.where(SBEntityMapping.EntityId == entity_id)

            result = await self.conn.execute(sb_sel_stmt)
            sb_ids = [row.SBId for row in result]

            del_stmt = delete(SBEntityMapping).where(
                SBEntityMapping.EntityId == entity_id
            )
            await self.conn.execute(del_stmt)

            sb_entity_sel_stmt = select(SBEntityMapping.SBId).where(
                SBEntityMapping.SBId.in_(sb_ids)
            )
            result = await self.conn.execute(sb_entity_sel_stmt)
            remaining_sb_ids = [row.SBId for row in result]
            if not remaining_sb_ids:
                unassign_stmt = (
                    update(SandBoxes)
                    .where(SandBoxes.SBId.in_(sb_ids))
                    .values(Assigned=False)
                )
                await self.conn.execute(unassign_stmt)

    @asynccontextmanager
    async def delete_unused_sandboxes(
        self, *, limit: int | None = None
    ) -> AsyncGenerator[AsyncGenerator[str, None], None]:
        """Get the sandbox PFNs to delete.

        The result of this function can be used as an async context manager
        to yield the PFNs of the sandboxes to delete. The context manager
        will automatically remove the sandboxes from the database upon exit.

        Args:
            limit: If not None, the maximum number of sandboxes to delete.

        """
        conditions = [
            # If it has assigned to a job but is no longer mapped it can be removed
            and_(
                SandBoxes.Assigned,
                ~exists().where(SBEntityMapping.SBId == SandBoxes.SBId),
            ),
            # If the sandbox is still unassigned after 15 days, remove it
            and_(~SandBoxes.Assigned, days_since(SandBoxes.LastAccessTime) >= 15),
        ]
        # Sandboxes which are not on S3 will be handled by legacy DIRAC
        condition = and_(SandBoxes.SEPFN.like("/S3/%"), or_(*conditions))

        # Copy the in-flight rows to a temporary table
        await self.conn.run_sync(partial(self._temp_table.create, checkfirst=True))
        select_stmt = select(SandBoxes.SBId).where(condition)
        if limit:
            select_stmt = select_stmt.limit(limit)
        insert_stmt = insert(self._temp_table).from_select(["SBId"], select_stmt)
        await self.conn.execute(insert_stmt)

        try:
            # Select the sandbox PFNs from the temporary table and yield them
            select_stmt = select(SandBoxes.SEPFN).join(
                self._temp_table, self._temp_table.c.SBId == SandBoxes.SBId
            )

            async def yield_pfns() -> AsyncGenerator[str, None]:
                async for row in await self.conn.stream(select_stmt):
                    yield row.SEPFN

            yield yield_pfns()

            # Delete the sandboxes from the main table
            delete_stmt = delete(SandBoxes).where(
                SandBoxes.SBId.in_(select(self._temp_table.c.SBId))
            )
            result = await self.conn.execute(delete_stmt)
            logger.info("Deleted %d expired/unassigned sandboxes", result.rowcount)

        finally:
            await self.conn.run_sync(partial(self._temp_table.drop, checkfirst=True))



# The utils class define some boilerplate types that should be used
# in place of the SQLAlchemy one. Have a look at them
from __future__ import annotations

from sqlalchemy import ForeignKey, Integer, String, Uuid
from sqlalchemy.orm import declarative_base

from diracx.db.sql.utils import Column, DateNowColumn

Base = declarative_base()


class Owners(Base):
    __tablename__ = "Owners"
    owner_id = Column("OwnerID", Integer, primary_key=True, autoincrement=True)
    creation_time = DateNowColumn("CreationTime")
    name = Column("Name", String(255))


class Cars(Base):
    __tablename__ = "Cars"
    license_plate = Column("LicensePlate", Uuid(), primary_key=True)
    model = Column("Model", String(255))
    owner_id = Column("OwnerID", Integer, ForeignKey(Owners.owner_id))



from __future__ import annotations

from sqlalchemy import (
    BigInteger,
    Boolean,
    Float,
    ForeignKey,
    Index,
    Integer,
    String,
)
from sqlalchemy.orm import declarative_base

from ..utils import Column

TaskQueueDBBase = declarative_base()


class TaskQueues(TaskQueueDBBase):
    __tablename__ = "tq_TaskQueues"
    TQId = Column(Integer, primary_key=True)
    Owner = Column(String(255), nullable=False)
    OwnerGroup = Column(String(32), nullable=False)
    VO = Column(String(32), nullable=False)
    CPUTime = Column(BigInteger, nullable=False)
    Priority = Column(Float, nullable=False)
    Enabled = Column(Boolean, nullable=False, default=0)
    __table_args__ = (Index("TQOwner", "Owner", "OwnerGroup", "CPUTime"),)


class JobsQueue(TaskQueueDBBase):
    __tablename__ = "tq_Jobs"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    JobId = Column(Integer, primary_key=True)
    Priority = Column(Integer, nullable=False)
    RealPriority = Column(Float, nullable=False)
    __table_args__ = (Index("TaskIndex", "TQId"),)


class SitesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToSites"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("SitesTaskIndex", "TQId"),
        Index("SitesIndex", "Value"),
    )


class GridCEsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToGridCEs"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("GridCEsTaskIndex", "TQId"),
        Index("GridCEsValueIndex", "Value"),
    )


class BannedSitesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToBannedSites"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("BannedSitesTaskIndex", "TQId"),
        Index("BannedSitesValueIndex", "Value"),
    )


class PlatformsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToPlatforms"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("PlatformsTaskIndex", "TQId"),
        Index("PlatformsValueIndex", "Value"),
    )


class JobTypesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToJobTypes"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("JobTypesTaskIndex", "TQId"),
        Index("JobTypesValueIndex", "Value"),
    )


class TagsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToTags"
    TQId = Column(
        Integer, ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value = Column(String(64), primary_key=True)
    __table_args__ = (
        Index("TagsTaskIndex", "TQId"),
        Index("TagsValueIndex", "Value"),
    )



from __future__ import annotations

from datetime import UTC, datetime

from diracx.db.os.utils import BaseOSDB


class JobParametersDB(BaseOSDB):
    fields = {
        "JobID": {"type": "long"},
        "timestamp": {"type": "date"},
        "PilotAgent": {"type": "keyword"},
        "Pilot_Reference": {"type": "keyword"},
        "JobGroup": {"type": "keyword"},
        "CPUNormalizationFactor": {"type": "long"},
        "NormCPUTime(s)": {"type": "long"},
        "Memory(MB)": {"type": "long"},
        "LocalAccount": {"type": "keyword"},
        "TotalCPUTime(s)": {"type": "long"},
        "PayloadPID": {"type": "long"},
        "HostName": {"type": "text"},
        "GridCE": {"type": "keyword"},
        "CEQueue": {"type": "keyword"},
        "BatchSystem": {"type": "keyword"},
        "ModelName": {"type": "keyword"},
        "Status": {"type": "keyword"},
        "JobType": {"type": "keyword"},
    }
    # TODO: Does this need to be configurable?
    index_prefix = "job_parameters"

    def index_name(self, vo, doc_id: int) -> str:
        split = int(int(doc_id) // 1e6)
        # The index name must be lowercase or opensearchpy will throw.
        return f"{self.index_prefix}_{vo.lower()}_{split}m"

    def upsert(self, vo, doc_id, document):
        document = {
            "JobID": doc_id,
            "timestamp": int(datetime.now(tz=UTC).timestamp() * 1000),
            **document,
        }
        return super().upsert(vo, doc_id, document)



from __future__ import annotations

from datetime import datetime
from functools import partial
from zoneinfo import ZoneInfo

import sqlalchemy.types as types
from sqlalchemy import Column as RawColumn
from sqlalchemy import DateTime, Enum

from .functions import utcnow

Column: partial[RawColumn] = partial(RawColumn, nullable=False)
NullColumn: partial[RawColumn] = partial(RawColumn, nullable=True)
DateNowColumn = partial(Column, type_=DateTime(timezone=True), server_default=utcnow())


def EnumColumn(name, enum_type, **kwargs):  # noqa: N802
    return Column(name, Enum(enum_type, native_enum=False, length=16), **kwargs)


class EnumBackedBool(types.TypeDecorator):
    """Maps a ``EnumBackedBool()`` column to True/False in Python."""

    impl = types.Enum("True", "False", name="enum_backed_bool")
    cache_ok = True

    def process_bind_param(self, value, dialect) -> str:
        if value is True:
            return "True"
        elif value is False:
            return "False"
        else:
            raise NotImplementedError(value, dialect)

    def process_result_value(self, value, dialect) -> bool:
        if value == "True":
            return True
        elif value == "False":
            return False
        else:
            raise NotImplementedError(f"Unknown {value=}")


class SmarterDateTime(types.TypeDecorator):
    """A DateTime type that also accepts ISO8601 strings.

    Takes into account converting timezone aware datetime objects into
    naive form and back when needed.

    """

    impl = DateTime()
    cache_ok = True

    def __init__(
        self,
        stored_tz: ZoneInfo | None = ZoneInfo("UTC"),
        returned_tz: ZoneInfo = ZoneInfo("UTC"),
        stored_naive_sqlite=True,
        stored_naive_mysql=True,
        stored_naive_postgres=False,  # Forces timezone-awareness
    ):
        self._stored_naive_dialect = {
            "sqlite": stored_naive_sqlite,
            "mysql": stored_naive_mysql,
            "postgres": stored_naive_postgres,
        }
        self._stored_tz: ZoneInfo | None = stored_tz  # None = Local timezone
        self._returned_tz: ZoneInfo = returned_tz

    def _stored_naive(self, dialect):
        if dialect.name not in self._stored_naive_dialect:
            raise NotImplementedError(dialect.name)
        return self._stored_naive_dialect.get(dialect.name)

    def process_bind_param(self, value, dialect):
        if value is None:
            return None

        if isinstance(value, str):
            try:
                value: datetime = datetime.fromisoformat(value)
            except ValueError as err:
                raise ValueError(f"Unable to parse datetime string: {value}") from err

        if not isinstance(value, datetime):
            raise ValueError(f"Expected datetime or ISO8601 string, but got {value!r}")

        if not value.tzinfo:
            raise ValueError(
                f"Provided timestamp {value=} has no tzinfo -"
                " this is problematic and may cause inconsistencies in stored timestamps.\n"
                " Please always work with tz-aware datetimes / attach tzinfo to your datetime objects:"
                " e.g. datetime.now(tz=timezone.utc) or use datetime_obj.astimezone() with no arguments if you need to "
                "attach the local timezone to a local naive timestamp."
            )

        # Check that we need to convert the timezone to match self._stored_tz timezone:
        if self._stored_naive(dialect):
            # if self._stored_tz is None, we use our local/system timezone.
            stored_tz = self._stored_tz

            # astimezone converts to the stored timezone (local timezone if None)
            # replace strips the TZ info --> naive datetime object
            value = value.astimezone(tz=stored_tz).replace(tzinfo=None)

        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        if not isinstance(value, datetime):
            raise NotImplementedError(f"{value=} not a datetime object")

        if self._stored_naive(dialect):
            # Here we add back the tzinfo to the naive timestamp
            # from the DB to make it aware again.
            if value.tzinfo is None:
                # we are definitely given a naive timestamp, so handle it.
                # add back the timezone info if stored_tz is set
                if self._stored_tz:
                    value = value.replace(tzinfo=self._stored_tz)
                else:
                    # if stored as a local time, add back the system timezone info...
                    value = value.astimezone()
            else:
                raise ValueError(
                    f"stored_naive is True for {dialect.name=}, but the database engine returned "
                    "a tz-aware datetime. You need to check the SQLAlchemy model is consistent with the DB schema."
                )

        # finally, convert the datetime according to the "returned_tz"
        value = value.astimezone(self._returned_tz)

        # phew...
        return value
